/*
 * Copyright (c) 2016 ARM Limited
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "mps2_glcd.h"

#include "cmsis.h"
#include "Driver_SPI.h"
#include "mps2_glcd_fonts.h"
#include "smm_mps2.h"
#include "mps2_time.h"

#include "device_cfg.h"

#define GLCD_LANDSCAPE 1
#define GLCD_ROTATE180 1

/* GLCD MISC masks */
#define GLCD_CS_Msk     (1UL << 0)
#define GLCD_RESET_Msk  (1UL << 3)
#define GLCD_RS_Msk     (1UL << 4)
#define GLCD_RD_Msk     (1UL << 5)
#define GLCD_BL_Msk     (1UL << 6)

/* CS pin setting to 0 or 1 */
#define LCD_CS(x)  ((x) ? (MPS2_FPGAIO->MISC |= GLCD_CS_Msk) \
                        : (MPS2_FPGAIO->MISC &= ~GLCD_CS_Msk))

#define SPI_START  0x70  /* Start byte for SPI transfer */
#define SPI_RD     0x01  /* WR bit 1 within start */
#define SPI_WR     0x00  /* WR bit 0 within start */
#define SPI_DATA   0x02  /* RS bit 1 within start byte */
#define SPI_INDEX  0x00  /* RS bit 0 within start byte */

#define PADDING_VAL  {0,0}

#define SPI_SET_MASK(ws)  (ARM_SPI_MODE_MASTER | ARM_SPI_CPOL0_CPHA0 | \
                           ARM_SPI_DATA_BITS(ws));

/* Text and background color */
struct screen_cfg_t {
    uint16_t fg_color;      /* Forground color */
    uint16_t bg_color;      /* backgound color */
    uint16_t width;         /* Screen width */
    uint16_t height;        /* Screen height */
    uint8_t  is_landscape;  /* Landscape config */
    uint8_t  is_rotate180;  /* Rotate screen 180 */
    uint8_t  padding[2];    /* to keep 32 bits alignment */
};

static struct screen_cfg_t glcd_cfg = { MPS2_GLCD_BLACK,
                                        MPS2_GLCD_WHITE,
                                        MPS2_GLCD_WIDTH,
                                        MPS2_GLCD_HEIGHT,
                                        GLCD_LANDSCAPE,
                                        GLCD_ROTATE180,
                                        PADDING_VAL};

/* Import GLCD SPI */
extern ARM_DRIVER_SPI Driver_SPI0;

/* Anonimous struct that holds SPI CMSIS driver configuration and bitrate */
static struct {
    uint32_t mask;
    uint32_t bitrate;
} spi_curr_cfg;

/*
 * \brief Transfers 1 byte over the SPI
 *
 * \param[in] byte  Byte to be sent
 */
static inline void spi_tran_byte(uint8_t byte)
{
    (void)Driver_SPI0.Send(&byte, 1);
}

/*
 * \brief Transfers n bytes over the SPI
 *
 * \param[in] bytes  Array of bytes to be sent
 * \param[in] n  Number of bytes to be sent
 */
static inline void spi_tran_n_bytes(uint8_t *bytes, uint16_t n)
{
    (void)Driver_SPI0.Send(bytes, n);
}

/*
 * \brief Change the wordsize of the SPI device
 *
 * \param[in] wordsize  Word size to set
 */
static inline void spi_change_wordsize(uint16_t wordsize)
{
    spi_curr_cfg.mask = (ARM_SPI_MODE_MASTER | ARM_SPI_CPOL0_CPHA0 );
    spi_curr_cfg.mask |= ((wordsize & 0x3FUL) << ARM_SPI_DATA_BITS_Pos);
    (void)Driver_SPI0.Control(spi_curr_cfg.mask, spi_curr_cfg.bitrate);
}

/*
 * \brief Writes a command the GLCD controller
 *
 * \param[in] cmd  Command to be written
 */
static inline void write_cmd(uint8_t cmd)
{
    LCD_CS(0);
    /* Write : RS = 0, RW = 0 */
    spi_tran_byte(SPI_START | SPI_WR | SPI_INDEX);
    spi_tran_byte(0);
    spi_tran_byte(cmd);
    LCD_CS(1);
}

/*
 * \brief Writes data to the LCD controller
 *
 * \param[in] data  Data to be written
 */
static inline void write_dat(uint16_t data)
{
    LCD_CS(0);
    /* Write : RS = 1, RW = 0 */
    spi_tran_byte(SPI_START | SPI_WR | SPI_DATA);
    spi_tran_byte((data >> 8));
    spi_tran_byte((uint8_t)data);
    LCD_CS(1);
}

/*
 * \brief Starts of data writing to the LCD controller
 */
static inline void write_dat_start(void)
{
    LCD_CS(0);
    /* Write : RS = 1, RW = 0 */
    spi_tran_byte(SPI_START | SPI_WR | SPI_DATA);
}

/*
 * \brief Stops of data writing to the LCD controller
 */
static inline void write_dat_stop(void)
{
    LCD_CS(1);
}

/*
 * \brief Data writing to the LCD controller
 *
 * \param[in] data  Data to be written
 */
static inline void write_dat_only(uint16_t data)
{
    spi_tran_byte((uint8_t)(data >> 8));
    spi_tran_byte((uint8_t)data);
}

/*
 * \brief Write array of data to the LCD controller
 *
 * \param[in] data  Array of data to be written
 * \param[in] nmi   Number of items to be written
 */
static inline void write_dat_array(void *data, uint16_t nmi)
{
    spi_tran_n_bytes(data, nmi);
}

/*
 * \brief Writes a value to the to LCD register
 *
 * \param[in] reg  Register to be written
 * \param[in] val  Value to write to the register
 */
static inline void write_reg(uint8_t reg, uint16_t val)
{
    write_cmd(reg);
    write_dat(val);
}

/*
 * \brief Resets GLCD screen
 */
static void reset_glcd(void)
{
    MPS2_FPGAIO->MISC |=  (GLCD_CS_Msk |    /* deassert nCS0 */
                           GLCD_RESET_Msk); /* deassert Reset */

    MPS2_FPGAIO->MISC &= ~(GLCD_RS_Msk |    /* deassert RS */
                           GLCD_RD_Msk |    /* deassert RD */
                           GLCD_BL_Msk );   /* switch off backlight */

    MPS2_FPGAIO->MISC &= ~GLCD_RESET_Msk;   /* assert Reset */
    mps2_sleepus(1);
    MPS2_FPGAIO->MISC |=  GLCD_RESET_Msk;   /* deassert Reset */
    mps2_sleepus(10);
}

/*
 * \brief Sets draw window region to maximum
 */
static void glcd_win_max(void)
{
    mps2_glcd_set_win(0, 0, MPS2_GLCD_WIDTH, MPS2_GLCD_HEIGHT);
}

void mps2_glcd_init(void)
{
    (void)Driver_SPI0.Initialize(NULL);

    spi_curr_cfg.mask = SPI_SET_MASK(8);
    spi_curr_cfg.bitrate = DEFAULT_SPI_SPEED_HZ;

    reset_glcd();

    /* Driving ability settings */
    write_reg(0xEA, 0x00);  /* Power control internal used (1) */
    write_reg(0xEB, 0x20);  /* Power control internal used (2) */
    write_reg(0xEC, 0x0C);  /* Source control internal used (1) */
    write_reg(0xED, 0xC7);  /* Source control internal used (2) */
    write_reg(0xE8, 0x38);  /* Source output period Normal mode */
    write_reg(0xE9, 0x10);  /* Source output period Idle mode */
    write_reg(0xF1, 0x01);  /* RGB 18-bit interface ;0x0110 */
    write_reg(0xF2, 0x10);

    /* Adjust the Gamma Curve */
    write_reg(0x40, 0x01);
    write_reg(0x41, 0x00);
    write_reg(0x42, 0x00);
    write_reg(0x43, 0x10);
    write_reg(0x44, 0x0E);
    write_reg(0x45, 0x24);
    write_reg(0x46, 0x04);
    write_reg(0x47, 0x50);
    write_reg(0x48, 0x02);
    write_reg(0x49, 0x13);
    write_reg(0x4A, 0x19);
    write_reg(0x4B, 0x19);
    write_reg(0x4C, 0x16);

    write_reg(0x50, 0x1B);
    write_reg(0x51, 0x31);
    write_reg(0x52, 0x2F);
    write_reg(0x53, 0x3F);
    write_reg(0x54, 0x3F);
    write_reg(0x55, 0x3E);
    write_reg(0x56, 0x2F);
    write_reg(0x57, 0x7B);
    write_reg(0x58, 0x09);
    write_reg(0x59, 0x06);
    write_reg(0x5A, 0x06);
    write_reg(0x5B, 0x0C);
    write_reg(0x5C, 0x1D);
    write_reg(0x5D, 0xCC);

    /* Power voltage setting */
    write_reg(0x1B, 0x1B);
    write_reg(0x1A, 0x01);
    write_reg(0x24, 0x2F);
    write_reg(0x25, 0x57);
    write_reg(0x23, 0x88);

    /* Power on setting */
    write_reg(0x18, 0x36);  /* Internal oscillator frequency adj */
    write_reg(0x19, 0x01);  /* Enable internal oscillator */
    write_reg(0x01, 0x00);  /* Normal mode, no scrool */
    write_reg(0x1F, 0x88);  /* Power control 6 - DDVDH Off */
    mps2_sleepus(20);
    write_reg(0x1F, 0x82);  /* Power control 6 - Step-up: 3 x VCI */
    mps2_sleepus(5);
    write_reg(0x1F, 0x92);  /* Power control 6 - Step-up: On */
    mps2_sleepus(5);
    write_reg(0x1F, 0xD2);  /* Power control 6 - VCOML active */
    mps2_sleepus(5);

    /* glcd_cfg selection */
    write_reg(0x17, 0x55);  /* RGB, System interface: 16 Bit/Pixel */
    write_reg(0x00, 0x00);  /* Scrolling off, no standby */

    /* Interface config */
    write_reg(0x2F, 0x11);  /* LCD Drive: 1-line inversion */
    write_reg(0x31, 0x00);
    write_reg(0x32, 0x00);  /* DPL=0, HSPL=0, VSPL=0, EPL=0 */

    /* Display on setting */
    write_reg(0x28, 0x38);  /* PT(0,0) active, VGL/VGL */
    mps2_sleepus(20);
    write_reg(0x28, 0x3C);  /* Display active, VGL/VGL */

    if(glcd_cfg.is_landscape) {
        if(glcd_cfg.is_rotate180) {
            write_reg (0x16, 0x68);
        } else {
            write_reg (0x16, 0xA8);
        }
    } else {
        if(glcd_cfg.is_rotate180) {
            write_reg (0x16, 0xC8);
        } else {
            write_reg (0x16, 0x08);
        }
    }

    /* Display scrolling settings  */
    write_reg(0x0E, 0x00);           /* TFA MSB */
    write_reg(0x0F, 0x00);           /* TFA LSB */
    write_reg(0x10, (320 >> 8));     /* VSA MSB */
    write_reg(0x11, (320 &  0xFF));  /* VSA LSB */
    write_reg(0x12, 0x00);           /* BFA MSB */
    write_reg(0x13, 0x00);           /* BFA LSB */

    /* Switch on backlight */
    MPS2_FPGAIO->MISC |= (GLCD_BL_Msk);
}

void mps2_glcd_set_win(uint16_t x, uint16_t y, uint16_t w, uint16_t h)
{
    uint16_t xe, ye;

    xe = x+w-1;
    ye = y+h-1;

    write_reg(0x02, (x  >> 8));      /* Column address start MSB */
    write_reg(0x03, (x & 0x00FF));   /* Column address start LSB */
    write_reg(0x04, (xe >> 8));      /* Column address end MSB */
    write_reg(0x05, (xe & 0x00FF));  /* Column address end LSB */

    write_reg(0x06, (y  >> 8));      /* Row address start MSB */
    write_reg(0x07, (y & 0x00FF));   /* Row address start LSB */
    write_reg(0x08, (ye >> 8));      /* Row address end MSB */
    write_reg(0x09, (ye & 0x00FF));  /* Row address end LSB */
}

void mps2_set_landscape(bool landscape)
{
    glcd_cfg.is_landscape = (uint8_t)landscape;

    if(glcd_cfg.is_landscape) {
        if(glcd_cfg.is_rotate180) {
            write_reg (0x16, 0x68);
        } else {
            write_reg (0x16, 0xA8);
        }
    } else {
        if(glcd_cfg.is_rotate180) {
            write_reg (0x16, 0xC8);
        } else {
            write_reg (0x16, 0x08);
        }
    }
}

void mps2_set_rotate180(bool rotate)
{
    glcd_cfg.is_rotate180 = (uint8_t)rotate;

    if(glcd_cfg.is_landscape) {
        if(glcd_cfg.is_rotate180) {
            write_reg (0x16, 0x68);
        } else {
            write_reg (0x16, 0xA8);
        }
    } else {
        if(glcd_cfg.is_rotate180) {
            write_reg (0x16, 0xC8);
        } else {
            write_reg (0x16, 0x08);
        }
    }
}

void mps2_glcd_put_pixel(uint16_t x, uint16_t y)
{
    write_reg(0x02, (x >> 8));      /* Column address start MSB */
    write_reg(0x03, (x & 0x00FF));  /* Column address start LSB */
    write_reg(0x04, (x >> 8));      /* Column address end MSB */
    write_reg(0x05, (x & 0x00FF));  /* Column address end LSB */

    write_reg(0x06, (y >> 8));      /* Row address start MSB */
    write_reg(0x07, (y & 0x00FF));  /* Row address start LSB */
    write_reg(0x08, (y >> 8));      /* Row address end MSB */
    write_reg(0x09, (y & 0x00FF));  /* Row address end LSB */

    write_reg(0x22, glcd_cfg.fg_color);
}

void mps2_glcd_put_pixel_color(uint16_t x, uint16_t y, uint16_t color)
{
    write_reg(0x02, (x >> 8));      /* Column address start MSB */
    write_reg(0x03, (x & 0x00FF));  /* Column address start LSB */
    write_reg(0x04, (x >> 8));      /* Column address end MSB */
    write_reg(0x05, (x & 0x00FF));  /* Column address end LSB */

    write_reg(0x06, (y >> 8));      /* Row address start MSB */
    write_reg(0x07, (y & 0x00FF));  /* Row address start LSB */
    write_reg(0x08, (y >> 8));      /* Row address end MSB */
    write_reg(0x09, (y & 0x00FF));  /* Row address end LSB */

    write_reg(0x22, color);
}

void mps2_glcd_set_text_color(uint16_t color)
{
    glcd_cfg.fg_color = color;
}

void mps2_glcd_set_back_color(uint16_t color)
{
    glcd_cfg.bg_color = color;
}

void mps2_glcd_clear(uint16_t color)
{
    uint32_t i;
    uint32_t screen_size;

    screen_size = glcd_cfg.width * glcd_cfg.height;

    glcd_win_max();

    write_cmd(0x22);
    write_dat_start();

    for(i = 0; i < screen_size; i++) {
        write_dat_only(color);
    }

    write_dat_stop();
}

void mps2_glcd_draw_char(uint16_t x, uint16_t y, uint16_t cw,
                         uint16_t ch, const uint8_t* c)
{
    uint16_t i, j, k, pixs;


    mps2_glcd_set_win(x, y, cw, ch);

    write_cmd(0x22);
    write_dat_start();

    k  = (cw + 7)/8;

    if (k == 1) {
        for (j = 0; j < ch; j++) {
          pixs = *c;
          c += 1;
          for (i = 0; i < cw; i++) {
              if(((pixs >> i) & 1)) {
                  write_dat_only(glcd_cfg.fg_color);
              } else {
                  write_dat_only(glcd_cfg.bg_color);
              }
          }
        }
    } else if (k == 2) {
        for (j = 0; j < ch; j++) {
          pixs = ((uint16_t)(*c << 8) + *c);
          c += 2;
          for (i = 0; i < cw; i++) {
              if(((pixs >> i) & 1)) {
                  write_dat_only(glcd_cfg.fg_color);
              } else {
                  write_dat_only(glcd_cfg.bg_color);
              }
          }
        }
    }

    write_dat_stop();
}


void mps2_glcd_display_char(uint16_t ln, uint16_t col,
                            enum mps2_glcd_font_size_t font, uint8_t c)
{
    /* Substact font offset */
    c -= 32;

    switch(font) {
    case MPS2_GLCD_FONT_SIZE_6x8:
        mps2_glcd_draw_char(col *  6, ln *  8,  6,  8,
                            &MPS2_GLCD_FONT_6x8[c*8]);
        break;
    case MPS2_GLCD_FONT_SIZE_16x24:
        mps2_glcd_draw_char(col * 16, ln * 24, 16, 24,
                            (const uint8_t*)(&MPS2_GLCD_FONT_16x24[c*24]));
        break;
    /* default:  The default is not defined intentionally to force the
     *           compiler to check that all the enumeration values are
     *           covered in the switch.*/
     }
}

enum mps2_glcd_error_t mps2_glcd_display_string(uint16_t ln, uint16_t col,
                                                enum mps2_glcd_font_size_t font,
                                                const char *p_str)
{
    if(p_str == NULL) {
        return MPS2_GLCD_ERR_INVALID_ARG;
    }

    while(*p_str) {
        mps2_glcd_display_char(ln, col++, font, *p_str++);
    }

    return MPS2_GLCD_ERR_NONE;
}

void mps2_glcd_clear_ln(uint16_t ln, enum mps2_glcd_font_size_t font)
{
    char buf[60];
    uint8_t i;

    glcd_win_max();

    switch(font) {
    case MPS2_GLCD_FONT_SIZE_6x8:
        for(i = 0; i < (glcd_cfg.width+5)/6; i++) {
            buf[i] = ' ';
        }
        buf[i+1] = 0;
        break;
    case MPS2_GLCD_FONT_SIZE_16x24:
        for(i = 0; i < (glcd_cfg.width+15)/16; i++) {
            buf[i] = ' ';
        }
        buf[i+1] = 0;
        break;
    /* default:  The default is not defined intentionally to force the
     *           compiler to check that all the enumeration values are
     *           covered in the switch.*/
    }

    mps2_glcd_display_string(ln, 0, font, buf);
}

void mps2_glcd_bargraph(uint16_t x,  uint16_t y, uint16_t w,
                        uint16_t h, uint16_t val)
{
    uint16_t i,j;
    uint32_t scale_val;

    scale_val = ((val * w) >> 10);  /* Scale value */

    mps2_glcd_set_win(x, y, w, h);

    write_cmd(0x22);
    write_dat_start();
    for (i = 0; i < h; i++) {
        for (j = 0; j <= w-1; j++) {
            if(j >= scale_val) {
                write_dat_only(glcd_cfg.bg_color);
            } else {
                write_dat_only(glcd_cfg.fg_color);
            }
        }
    }
    write_dat_stop();
}

enum mps2_glcd_error_t mps2_glcd_bitmap(uint16_t x, uint16_t y, uint16_t w,
                                        uint16_t h, uint16_t* bitmap)
{
    uint32_t i;

    if(bitmap == NULL) {
        return MPS2_GLCD_ERR_INVALID_ARG;
    }

    mps2_glcd_set_win(x, y, w, h);

    write_cmd(0x22);
    write_dat_start();

    /* Send whole row in every iteration, with items of 16 bits */
    spi_change_wordsize(16);
    for(i = 0; i < (w*h); i+=w) {
        write_dat_array(&bitmap[i], w);
    }
    spi_change_wordsize(8);

    write_dat_stop();

    return MPS2_GLCD_ERR_NONE;
}

void mps2_glcd_scroll_vertical(uint16_t dy)
{
    uint16_t y = dy;

    if(glcd_cfg.is_landscape) {
        while(y >= glcd_cfg.height) {
            y -= glcd_cfg.height;
        }

        write_reg(0x01, 0x08);
        write_reg(0x14, (y >> 8));    /* VSP MSB */
        write_reg(0x15, (y & 0x00FF));  /* VSP LSB */
    }
}

void mps2_glcd_box(uint16_t x, uint16_t y, uint16_t w,
                   uint16_t h, uint16_t color)
{
    uint32_t i;

    mps2_glcd_set_win(x, y, w, h);

    write_cmd(0x22);
    write_dat_start();
    for(i = 0; i < (w*h); i++) {
        write_dat_only(color);
    }
    write_dat_stop();
}

void mps2_glcd_box_border(uint16_t x, uint16_t y, uint16_t w,
                          uint16_t h, uint16_t color)
{
    uint16_t i;
    uint16_t xe, ye;

    xe = x+w;
    ye = y+h;

    for(i = y; i < ye; i++) {
        mps2_glcd_put_pixel_color(x, i, color);
        mps2_glcd_put_pixel_color(xe, i, color);
    }

    for(i = x; i < xe; i++) {
        mps2_glcd_put_pixel_color(i, y, color);
        mps2_glcd_put_pixel_color(i, ye, color);
    }
}
