/*
 * Copyright (c) 2016 ARM Limited
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef __S_NS_SHARED_H__
#define __S_NS_SHARED_H__

#ifdef __cplusplus
extern "C" {
#endif

#include <stdint.h>
#include <stdbool.h>

/*
 * This function pointer is meant to only hold non secure function pointers.
 * It will be turned into a non-secure one (LSB cleared) before being called
 * whatever happens anyway (unless cast to another function pointer type).
 * Registers will be cleared before branching so that no information leaks
 * from secure to non-secure world.
 */
typedef void __attribute__((cmse_nonsecure_call)) (*nsfptr_t) (void);

/* MPS2 user button mask */
#define BT0_MASK  0x00000001
#define BT1_MASK  0x00000002

/*
 * Those functions are an entry point for the non-secure world. That means
 * that the non-secure world can call this function through the veneer placed
 * in a non-secure callable area.
 * If the secure world tries to take the address of this function,
 * it may get the address of the veneer or the adress of the function
 * itself, depending on linker implementation.
 * If the non-secure world tries to take the address of this function,
 * it will always get the address of the veneer (as defined in the
 * s_veneers.o generated with --import_cmse_lib_out on armclang).
 */

/**
 * \brief  Gets MPS2 buttons status.
 *
 * \return Returns MPS2 buttons status (0x0000SSBB) where MCC switchs
 *         status and BB user button status.
 */
extern uint32_t get_btns_state(void)
       __attribute__((cmse_nonsecure_entry, noinline));

/**
 * \brief      Sleeps X amount of ms.
 *
 * \paramp[in] ms  Number of miliseconds to sleep
 */
extern void sleep_ms(uint32_t ms)
       __attribute__((cmse_nonsecure_entry, noinline));

#ifdef __cplusplus
}
#endif
#endif /* __S_NS_SHARED_H__ */
