/*
 * Copyright (c) 2016-2017 ARM Limited
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "apb_timer_drv.h"

#include <stddef.h>

/* Timer state definitions */
#define APB_TIMER_INITIALIZED  (1 << 0)
#define APB_TIMER_ENABLED      (1 << 1)

/* Timer ctrl register definitions */
#define APB_TIMER_CTRL_IRQ_EN       (0x01ul << 3)
#define APB_TIMER_CTRL_SEL_EXT_CLK  (0x01ul << 2)
#define APB_TIMER_CTRL_SEL_EXT_EN   (0x01ul << 1)
#define APB_TIMER_CTRL_EN           (0x01ul << 0)
#define APB_TIMER_CTRL_INT_CLEAR    (0x01ul << 0)

/* Timer register map structure */
struct apb_timer_reg_map_t {
    volatile uint32_t ctrl;    /* Offset: 0x000 (R/W) control register */
    volatile uint32_t value;   /* Offset: 0x004 (R/W) current value register */
    volatile uint32_t reload;  /* Offset: 0x008 (R/W) reload value register */
    union {
        volatile uint32_t intstatus;  /* Offset: 0x00C (R/ ) interrupt
                                       * status register */
        volatile uint32_t intclear;   /* Offset: 0x00C ( /W) interrupt
                                       * clear register */
    }intreg;
};

void apb_timer_init(struct apb_timer_dev_t* dev, uint32_t load_val)
{
    uint32_t reload = APB_TIMER_DEFAULT_RELOAD;

    if(load_val != 0) {
        reload = load_val;
    }

    dev->data->reload = reload;
    dev->data->state = APB_TIMER_INITIALIZED;
}

uint32_t apb_timer_is_enabled(struct apb_timer_dev_t* dev)
{
    /* Timer has to be already initialized and enabled */
    return (dev->data->state & APB_TIMER_ENABLED);
}

enum apb_timer_error_t apb_timer_enable(struct apb_timer_dev_t* dev)
{
    struct apb_timer_reg_map_t* p_timer =
                                   (struct apb_timer_reg_map_t*)dev->cfg->base;

    /* Verify if the timer is initialized */
    if(!(dev->data->state & APB_TIMER_INITIALIZED)) {
        return APB_TIMER_ERR_NOT_INIT;
    }

    /* Disable timer */
    p_timer->ctrl = 0x0;

    /* Reload value */
    p_timer->reload = dev->data->reload;

    /* Enable counter */
    p_timer->ctrl = APB_TIMER_CTRL_EN;

    /* Change timer state */
    dev->data->state |= APB_TIMER_ENABLED;

    return APB_TIMER_ERR_NONE;
}

void apb_timer_disable(struct apb_timer_dev_t* dev)
{
    struct apb_timer_reg_map_t* p_timer =
                                   (struct apb_timer_reg_map_t*)dev->cfg->base;

    if(apb_timer_is_enabled(dev)) {
        /* Disable timer */
        p_timer->ctrl = 0x0;

        /* Change timer state */
        dev->data->state = APB_TIMER_INITIALIZED;
    }
}

uint32_t apb_timer_read(struct apb_timer_dev_t* dev)
{
    uint32_t return_value = 0;
    struct apb_timer_reg_map_t* p_timer;

    if(apb_timer_is_enabled(dev)) {
        p_timer = (struct apb_timer_reg_map_t*)dev->cfg->base;

        return_value = (dev->data->reload - p_timer->value);
    }

    return return_value;
}

enum apb_timer_error_t apb_timer_enable_interrupt(
                                          struct apb_timer_dev_t* dev,
                                          uint32_t load_val)
{
    struct apb_timer_reg_map_t* p_timer =
                                   (struct apb_timer_reg_map_t*)dev->cfg->base;

    /* Verify if the timer is initialized */
    if(!(dev->data->state & APB_TIMER_INITIALIZED)) {
        return APB_TIMER_ERR_NOT_INIT;
    }

    /* Disable timer */
    p_timer->ctrl = 0x0;

    /* Enable interrupt */
    p_timer->ctrl = APB_TIMER_CTRL_IRQ_EN;

    /* Check time us condition */
    if(load_val == 0) {
        load_val = APB_TIMER_DEFAULT_RELOAD;
    }

    /* Initialize timer value */
    dev->data->reload = load_val;
    p_timer->reload = load_val;
    p_timer->value = load_val;

    /* Enable counter */
    p_timer->ctrl |= APB_TIMER_CTRL_EN;

    /* Change timer state */
    dev->data->state |= APB_TIMER_ENABLED;

    return APB_TIMER_ERR_NONE;
}

void apb_timer_disable_interrupt(struct apb_timer_dev_t* dev)
{
    struct apb_timer_reg_map_t* p_timer =
                                   (struct apb_timer_reg_map_t*)dev->cfg->base;

    if(apb_timer_is_enabled(dev)) {
        /* Disable interrupt */
        p_timer->ctrl &= APB_TIMER_CTRL_EN;
    }
}

void apb_timer_clear_interrupt(struct apb_timer_dev_t* dev)
{
    struct apb_timer_reg_map_t* p_timer =
                                   (struct apb_timer_reg_map_t*)dev->cfg->base;

    /* Clear interrupt */
    p_timer->intreg.intclear = APB_TIMER_CTRL_INT_CLEAR;
}

uint32_t apb_timer_get_status(struct apb_timer_dev_t* dev)
{
    struct apb_timer_reg_map_t* p_timer =
                                   (struct apb_timer_reg_map_t*)dev->cfg->base;

    if(apb_timer_is_enabled(dev)) {
        return p_timer->intreg.intstatus;
    }
    return 0;
}

uint32_t apb_timer_get_reload_value(struct apb_timer_dev_t* dev)
{
    if(apb_timer_is_enabled(dev)) {
        return dev->data->reload;
    }
    return 0;
}
